// src/backend/scheduler.cc - Copyright 2005, 2006, University
//               of Padova, dept. of Pure and Applied
//               Mathematics
//
// This file is part of SGPEMv2.
//
// This is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// SGPEMv2 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with SGPEMv2; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

#include "concrete_environment.hh"
#include "concrete_history.hh"
#include "cpu_policy.hh"
#include "scheduler.hh"
#include "user_interrupt_exception.hh"

// Do not include full template definition in the header file
#include "singleton.tcc"

#include <glibmm/thread.h>

#include <cassert>
#include <memory>

using namespace std;
using namespace sgpem;

// Explicit template instantiation to allow to export symbols from the DSO.
template class SG_DLLEXPORT Singleton<Scheduler>;


typedef std::vector<DynamicProcess*>    Processes;
typedef std::vector<DynamicRequest*>    Requests;
typedef std::vector<DynamicSubRequest*> SubRequests;
typedef std::vector<DynamicThread*>     Threads;


// ------------------ Static helper functions --------------


// Collects all threads of an environment into a single vector
static void 
collect_threads(const std::vector<Process*>& procs, 
		std::vector<DynamicThread*>& collected_threads)
{
  typedef std::vector<Process*> Processes;
  typedef std::vector<DynamicThread*>  Threads;

  collected_threads.clear();
  for(Processes::const_iterator it1 = procs.begin(); it1 != procs.end(); it1++)
    {
      const Threads& ts = ((DynamicProcess&) **it1).get_dynamic_threads();
      collected_threads.insert(collected_threads.end(), ts.begin(), ts.end());
    }
}

static void prepare_ready_queue(ConcreteEnvironment&         snapshot, 
																std::vector<DynamicThread*>& all_threads)
{
	typedef std::vector<DynamicThread*> Threads;
	ReadyQueue& queue = snapshot.get_sorted_queue();
	assert(queue.size() == 0);
	for(Threads::const_iterator it = all_threads.begin(); 
			it != all_threads.end(); it++)
		{
			if((*it)->get_state() == Schedulable::state_ready)
				queue.append(**it);
		}
}


// For the current thread, see if there are requests that are exhausted
// see extendRequest, case 0 and 1
// static void
// update_requests_for_old_running_thread(DynamicThread& running_thread)
// {
// }





  /// \brief Manages a single SubRequest object, depending on its state.
	/// Zero step: any -> terminated. Added to cope with malformed threads.
  /// First step: allocated -> terminated.
  /// Second step: non-allocable -> allocable.
  /// Third step: allocable -> allocated, or future -> allocated.
  ///
  /// The name and the structure of this method are ugly. They are inherited
  /// from the whole visitor's structure, anyway we could simply switch on of
  /// the state the SubRequest obejct, and we could (should?) factor out the
  /// operations which check if the request is allocable or not, depending on
  /// the queue position. Anyway, this factoring may lead to code dispersion.
  /// I think it is better to hold the computational core in one single place.
  void
	extendSubRequest(DynamicSubRequest* sp, auto_ptr<ConcreteEnvironment> & env, int walk, int front)
  {
		DynamicSubRequest& s = *sp;
    switch (walk)
    {
    /// Terminates directly the subrequest
    case 0:
      {
        s.set_state(Request::state_exhausted);
        /// Remove the subrequest (pointer) from the queue.
        bool found = false;
        typedef Environment::SubRequestQueue SubRequestQueue;
        SubRequestQueue& queue = env->get_request_queue(s.get_resource_key());
        SubRequestQueue::iterator it = queue.begin();
        for (; !found && it != queue.end(); it++)
          if ((*it) == sp)
          {
            found = true;
            queue.erase(it);
          }
        break;
      }
    /// Updates the state of an ALLOCATED subrequest, decreasing appropriate
    /// counters, and checks if it become TERMINATED. In the latter case the
    /// function finds the position of the subrequest (pointer) in the
    /// requested resource's queue and removes it.
    case 1:
      {
        if (s.get_state() != Request::state_allocated)
          break;
        /// Decrease remaining time, since the resource has been used.
        s.decrease_remaining_time();
        /// Check for termination.
        if (s.get_remaining_time() == 0)
        {
          s.set_state(Request::state_exhausted);
          /// Remove the subrequest (pointer) from the queue.
          bool found = false;
					typedef Environment::SubRequestQueue SubRequestQueue;
					SubRequestQueue& queue = env->get_request_queue(s.get_resource_key());
          SubRequestQueue::iterator it = queue.begin();
          for (; !found && it != queue.end(); it++)
					{
            if ((*it) == sp)
            {
              found = true;
              queue.erase(it);
            }
					}
        }
        break;
      }
/*
    /// Updates the state of a NON-ALLOCABLE subrequest, in case it become
    /// ALLOCABLE, which may happen only when a resource has been released.
    /// We could jump this check if no resource were released.
    /// It finds the position of the subrequest (pointer) in the requested
    /// resource's queue. If the position is within the places of the
    /// resource, the subrequest is ALLOCABLE.
    case 2:
      {
        if (s.get_state() != Request::state_allocated)
          break;
        /// The subrequest is in the queue for sure. Let's find it!
        /// I really need an integer for this operation.
        int position = 0;
        while (position <= s.resource_ptr->queue.size())
        {
          if (s.resource_ptr->queue[position]->has_same_id(s))
            /// Found!
            break;
          /// This statement is not executed if we find it.
          position++;
        }
        /// Watch out: in a resource with 2 places, 0 and 1 are valid queue
        /// positions, 2 is right one place out.
        if (position < s.resource_ptr->places)
          /// If so, set it ALLOCABLE.
          s.set_state("ALLOCABLE");
        break;
      }
    /// Updates the state of a FUTURE subrequest when the time has come
    /// for it to be raised, setting it as allocable it if it is the case,
    /// or blocking it. Enqueues the subrequest (pointer) at the end of the
    /// requested resource's queue.
    /// The resource policy should be called to manage the queue.
    /// If the position is within the places of the resource, the subrequest
    /// is ALLOCABLE, otherwise it is NON-ALLOCABLE.
    case 3:
      {
        if (s.get_state() != "FUTURE")
          break;
        /// Enqueue the subrequest at the back of the queue.
        s.resource_ptr->queue.push_back(&s);
        /// TODO: right here, right now we should call the resource policy to
        /// update the queue. Updates the state of the subrequest depending
        /// on the position in the queue, as explained before.
        s.set_state(s.resource_ptr->queue.size() > s.resource_ptr->places ?
          "NON-ALLOCABLE" : "ALLOCABLE");
        // Oh I miss ML so much.
        break;
      }
    /// This is ugly, but hey, none's perfect.
    /// Updates the state of a ALLOCABLE subrequest allocating it.
    case 4:
      {
        if (s.get_state() == "ALLOCABLE")
          s.set_state("ALLOCATED");
        break;
      }
*/
    }
  }




  /// \brief Manages a single Request object, depending on its state.
  /// Updates the state of a request, depending on its state, recursively
  /// updating the contained subrequests. The state of the request is then
  /// a function of the states of the subrequests.
  ///
	/// Zero step: any -> terminated. Added to cope with malformed threads.
  /// First step: allocated -> terminated.
  /// Second step: non-allocable -> allocable.
  /// Third step: allocable -> allocated, or future -> allocated.
  ///
  /// The following function may be reduced to a pair of lines.
  ///
  /// Longer, safer and more efficient version (and hopefully much easier
  /// to understand!)
  void
  extendRequest(DynamicRequest* rp, auto_ptr<ConcreteEnvironment> & env, int walk, int front)
  {
		DynamicRequest& r = *rp;
    switch (walk)
    {
    case 0:
      {
				typedef vector<DynamicSubRequest*> SubRequests;
				SubRequests list = r.get_dynamic_subrequests();
        for (SubRequests::iterator it = list.begin(); it != list.end(); it++)
					extendSubRequest(*it, env, walk, front);
        break;
      }
    /// Updates the state of an ALLOCATED request.
    case 1:
      {
        if (r.get_state() != Request::state_allocated)
          break;
				typedef vector<DynamicSubRequest*> SubRequests;
				SubRequests list = r.get_dynamic_subrequests();
        for (SubRequests::iterator it = list.begin(); it != list.end(); it++)
					extendSubRequest(*it, env, walk, front);
        break;
      }
/*
    /// Updates the state of a NON-ALLOCABLE request.
    case 2:
      {
        if (r.get_state() != "NON-ALLOCABLE")
          break;
        for (int j = 0; j < r.subrequests.size(); j++)
          r.subrequests[j].accept(*this);
        break;
      }
    /// Updates the state of an ALLOCABLE or FUTURE request.
    case 3:
      {
        /// This is the only interesting case. If the current instant, measured
        /// over the containing process execution time, is equal to the instant
        /// in which the request has to be raised, the subrequests are
        /// recursively updated for the first time ever, and their status
        /// changes from FUTURE to something else.
        if (r.get_state() == "FUTURE" && r.at == front)
          for (int j = 0; j < r.subrequests.size(); j++)
            r.subrequests[j].accept(*this);
        if (r.get_state() == "ALLOCABLE")
        {
          walk = 4; // this is an ugly patch please forgive me
          for (int j = 0; j < r.subrequests.size(); j++)
            r.subrequests[j].accept(*this);
          walk = 3;
        }
        break;
      }
*/
    }
  }







// ---------------------------------------------------------




//private constructor. The parameter is discarded
Scheduler::Scheduler()
	: _ready_queue(NULL), _policy(NULL), _step_mutex()
{
}


ReadyQueue*
Scheduler::get_ready_queue()
{
  return _ready_queue;
}


/** \note E' fondamentale che questo metodo memorizzi localmente qualora la politica
 attuale sia a prerilascio o meno, e la durata del quanto di tempo, in quanto la politica
 e' libera di variare questi parametri a piacere durante l'esecuzione della simulazione
*/
void
Scheduler::reset_status()
{
  // DEPRECATED (?)
}


CPUPolicy*
Scheduler::get_policy()
{
	return _policy;
}


bool
Scheduler::step_forward(History& history, CPUPolicy& cpu_policy) throw(UserInterruptException)
{
	// This very method should be exclusive: no concurrent behaviour, from when we 
	// store a readyqueue and policy pointer for the user-policy to retrieve, to when 
	// the policy returns
	// TODO: restrict this area to maximise parallelism
  Glib::Mutex::Lock lock(_step_mutex);

	// NOTE: Be sure to read the *ORIGINAL* documentation in the design document for this method!

	// FIXME: handle me! I'm not just a pretty boolean, I want to be *USED*! *EXPLOITED*!
	//        *RAPED*! *MAKE ME BLEED*!
	bool simulation_ended = true; // Assume we've finished. Then prove me wrong.
	unsigned int current_instant = history.get_size(); /* They should be equivalent */

	ConcreteHistory& concrete_history = (ConcreteHistory&) history;
 
	// Use an auto_ptr since we've some exceptions in the coming...
	auto_ptr<ConcreteEnvironment> new_snapshot(new ConcreteEnvironment(concrete_history.get_last_environment()));
 
	Threads all_threads;
	DynamicThread* running_thread = NULL;

	collect_threads(new_snapshot->get_processes(), all_threads);

	// designer + implementer (Matteo) comment follows:

	for(Threads::iterator it = all_threads.begin(); it != all_threads.end(); it++)
		{
			DynamicThread& current = **it; 
    
			// 1. mark future threads as ready, if appropriate
			if(current.get_state() == Schedulable::state_future)
				{
					Process& parent = current.get_process();
					if(parent.get_elapsed_time() == current.get_arrival_time())
						current.set_state(Schedulable::state_ready);
				}
           
			// Save the current running thread for future usage, if it hasn't ended
			// its allotted time
			if(current.get_state() == Schedulable::state_running)
				{
					assert(running_thread == NULL); // ... and one to bind them all.
					running_thread = &current; // Even if we change its state to terminate

					// increasing the time elapsed of the running thread + process 
					// should  be done here as the first thing, instead than 
					// directly after selecting them
					if(current.get_total_cpu_time() - current.get_elapsed_time() > 0)
						current.decrease_remaining_time();

					// 2. mark threads that used all their allotted time as terminated
					if(current.get_total_cpu_time() - current.get_elapsed_time() == 0)
						current.set_state(Schedulable::state_terminated);
				} 

			// 3. check for simulation termination (we can directly use threads
			// for this check, since processes' state is based upon threads' one)
			if( /* we still think that */ simulation_ended && 
					(current.get_state() & (Schedulable::state_blocked | 
																	Schedulable::state_terminated)) == 0)
				simulation_ended = false;
		}
 
	// 4a. Look for exhausted requests for the running thread
	if(running_thread != NULL) 
  {
    bool running_terminated = running_thread->get_state() == Schedulable::state_terminated;	  

    Requests& reqs = running_thread->get_dynamic_requests();
	  for(Requests::iterator r_it = reqs.begin(); r_it != reqs.end(); r_it++)
      extendRequest(*r_it, new_snapshot, running_terminated ? 0 : 1, running_thread->get_elapsed_time());
  }
	
	
	// ----------  FIXME ----------------
	// Check correctness: Now if the simulation ended we
  // append the newly created environment and return false
  if(simulation_ended) goto final_cleanup;
 

	/*                    /
	 *                  /
	 *                /
	 * (I'M HERE)   < * * * * * * * * * * *  
	 *								\
	 *	  							\
	 *                    \
	 *
	 * (is it visible enough for you?)
	 */

 
	bool preemptible_policy;
	unsigned int  time_slice;
	try
		{
			// Temporarily set the _ready_queue param and the _policy one for 
			// use from external plugins
			_policy = &cpu_policy;
			_ready_queue = &new_snapshot->get_sorted_queue();
			
			// ?. Use the policy to sort the queue

			preemptible_policy = cpu_policy.is_pre_emptive();
			time_slice = cpu_policy.get_time_slice();
			
			// ?. See if old running_thread has to be put to ready state
			// This happens when the policy makes use of preemptability by
			// priority, or when a time slice ended
			if(running_thread != NULL && running_thread->get_state() == Schedulable::state_running &&
				   (preemptible_policy ||
						time_slice == current_instant - running_thread->get_last_acquisition()) )
				{
					running_thread->set_state(Schedulable::state_ready);
					running_thread->set_last_release(current_instant);
				}



      // Try to continue running the current running thread
      if (running_thread != NULL && running_thread->get_state() == Schedulable::state_running)
      {
        // the thread may block on raising a request
        Requests& reqs = running_thread->get_dynamic_requests();
        for(Requests::iterator r_it = reqs.begin(); r_it != reqs.end(); r_it++)
        {
          DynamicRequest& rq = **r_it;
          if (rq.get_state() == Request::state_future
            && rq.get_instant() == running_thread->get_elapsed_time())
          {
            SubRequests& subreqs = rq.get_dynamic_subrequests();
            for(SubRequests::iterator s_it = subreqs.begin(); s_it != subreqs.end(); s_it++)
            {
              DynamicSubRequest& subr = **s_it;
							Environment::SubRequestQueue& queue = new_snapshot->get_request_queue(subr.get_resource_key());
							queue.push_back(*s_it);
							// FIXME this code has to control the position in the queue
              if(subr.get_state() == Request::state_future)
                subr.set_state(Request::state_allocated);
            }
          }
        }
      }


			// ?. Ask the policy to sort the queue. We do this multiple time if we must select
			// a new thread and it can't run for some reason (goes blocked, or terminates).
			bool found = true;
			do
				{
					found = true;
					prepare_ready_queue(*new_snapshot, all_threads);
					
					if(_ready_queue->size() == 0) 
						break; // No sense in trying to schedule something that isn't there
					
					cpu_policy.sort_queue();
					
					DynamicThread& candidate = (DynamicThread&) _ready_queue->get_item_at(0);
					

          // the thread may block on raising a request
          Requests& reqs = candidate.get_dynamic_requests();
          for(Requests::iterator r_it = reqs.begin(); r_it != reqs.end(); r_it++)
          {
            DynamicRequest& rq = **r_it;
            if (rq.get_state() == Request::state_future
              && rq.get_instant() == candidate.get_elapsed_time())
            {
              SubRequests& subreqs = rq.get_dynamic_subrequests();
              for(SubRequests::iterator s_it = subreqs.begin(); s_it != subreqs.end(); s_it++)
                {
                  DynamicSubRequest& subr = **s_it;
									Environment::SubRequestQueue& queue = new_snapshot->get_request_queue(subr.get_resource_key());
									queue.push_back(*s_it);
									// FIXME this code has to control the position in the queue
            		  if(subr.get_state() == Request::state_future)
                		subr.set_state(Request::state_allocated);
                }
            }
          }


					// We could have threads with 0 duration. Silly, but possible.
          // the silly thing was to permit the user to do this :P
					if(candidate.get_total_cpu_time() - candidate.get_elapsed_time() == 0)
						{
							candidate.set_last_acquisition(current_instant);
							candidate.set_last_release(current_instant);
							candidate.set_state(Schedulable::state_terminated);
            // FIXED : check requests for thread at instant 0?
            // the real question is: should we check for requests raised at
            // the thread last instant? the answer is: who cares?
            // doing it or not is a matter of cut-n-paste
							found = false;
							continue;
						}

					// FIXME : check if the first thread of the queue blocks
					
				}
			while(!found);

			// ?. Finally select the new thread (if appropriate); now we're sure it can run
			if(_ready_queue->size() > 0 && (running_thread == NULL || running_thread->get_state() != Schedulable::state_running))
				{
					// Fix fields of running thread
					DynamicThread& new_running = (DynamicThread&) _ready_queue->get_item_at(0);
					new_running.set_state(Schedulable::state_running);
					new_running.set_last_acquisition(current_instant);
          // removes running element from the ready queue
          // since no method was provided to erase an item in the queue, we should rebuild it.
          // this is pointless. I just added the method.
          // rebuilding the ready queue may corrupt the order set by the policy:
          // this is not acceptable, nor it is asking the policy to resort it.
          _ready_queue->erase_first();
				}
					

		}
	catch(UserInterruptException& e)
		{
			// Reset values that the policy doesn't need anymore
			_policy      = NULL;
			_ready_queue = NULL;

			//  Do we need to update/reset something else?

			// Going up unwinding the stack, tell:
			//    - the user that the policy sucks
			//    - SimulationController that everything stopped
			throw;
		}		

 final_cleanup:
 
	// append the new snapshot...
	// ...and remember to release the auto_ptr!
	concrete_history.append_new_environment(new_snapshot.release());

	// Reset values that the policy doesn't need anymore
	_policy      = NULL;
	_ready_queue = NULL;

  // If we got there, a step has been performed
  return simulation_ended == false;
}
















